//  GAMBIT: Global and Modular BSM Inference Tool
//  *********************************************
///  \file
///
///  Prior object to wrap a user-supplied
///  inverse transform function
///
///  *********************************************
///
///  Authors (add name and date if you modify):
///   
///  \author Ben Farmer
///          (b.farmer@imperial.ac.uk))
///  \date 2018 Dec
///
///  *********************************************

#include <sstream>
#include "gambit/ScannerBit/base_userfunc.hpp"
#include "gambit/ScannerBit/priors/user_supplied.hpp"
#include "gambit/Utils/stream_overloads.hpp"

namespace Gambit
{
    namespace Priors
    {
        UserFuncPrior::UserFuncPrior(const std::vector<std::string>& params, const Options& /*options*/)
         : BasePrior(params, params.size())
         , my_func()
         , user_func_set(false)
        {}
 
        // Function to set the user-supplied transform function
        void UserFuncPrior::set_transform_func(UserFunc const * const user_func)
        {
           my_func = user_func;
           user_func_set = true;
        }

        // Transformation from unit cube to the model parameter space 
        void UserFuncPrior::transform(const std::vector <double> &unitpars, std::unordered_map <std::string, double> &outputMap) const
        {
           // Perform user-defined transformation on our sub-cube of parameters
           std::vector<double> vec; 
           if(user_func_set)
           {
              vec = my_func->transform(unitpars);
           }
           else
           {
              std::stringstream err;
              err << "Attempted to perform prior 'transform' on parameters "<<param_names<<" using UserFuncPrior, however the user-supplied function was never set! Something has gone wrong with the construction of this prior, please file a bug report";
              Scanner::scan_error().raise(LOCAL_INFO, err.str());        
           }

           // The parameter ordering is defined by the order of parameter names provided to the UserFuncPrior constructor           
           auto v_it = vec.begin();
           for (auto str_it = param_names.begin(), str_end = param_names.end(); str_it != str_end; str_it++)
           {
              outputMap[*str_it] = *(v_it++);
           }
        }
 
        // I think this is the prior density. We don't actually use this so far, I think.           
        double UserFuncPrior::operator()(const std::vector<double>&) const
        { return -1.; }
    }
}
